/* ====================================================================
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 2000-2003 Justin Kirby
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Apache" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 */
#include "oapr/oapr.hpp"
#include <sstream>
#include "oapr/oapr_buildconfig.hpp"
#include "oapr/oapr_io_socket.hpp"
#include "oapr/oapr_exception.hpp"

FILE_VERSION("$Revision: 1.5.2.2 $",oapr_io_socket_cpp);
namespace oapr
{
  apr_status_t basic_socket::create(int family, int type)
  {
    if(is_created() || is_open()){
      OAPR_THROW_LVL(logic_error,
		     "socket is already initialized",exception_base::EXC_Warning);
    }
    apr_status_t stat;
    stat = apr_socket_create(&m_sock,family,type,m_pool->getp());
    if(stat==APR_SUCCESS){
      m_socktype=type;
      m_created=true;
    }
    return stat;
  }
  apr_status_t basic_socket::shutdown(apr_shutdown_how_e how)
  {
    apr_status_t stat(APR_SUCCESS);
    if(is_open()){
      if((stat=apr_socket_shutdown(m_sock,how))==APR_SUCCESS){
        m_isopen=false;
        m_created=false;
      }
    }
    return stat;
  }
  apr_status_t basic_socket::bind(const sockinfo& si)
  {
    if(!is_created()){
      OAPR_THROW_LVL(logic_error,
		     "socket is not created",exception_base::EXC_Error);
    }
    apr_status_t stat;
    stat=apr_socket_bind(m_sock,si.sockaddr());
    if(stat==APR_SUCCESS){
      m_isopen=true;
    }
    return stat;
  }
  apr_status_t basic_socket::listen(apr_int32_t backlog)
  {
    if(!is_open()){
      OAPR_THROW_LVL(logic_error,
		     "socket is not opened",exception_base::EXC_Crit);
    }
    /**\todo should this be more checked? or is this simple return ok?*/
    return apr_listen(m_sock,backlog);
  }
  basic_socket basic_socket::accept()
  {
    if(!is_open() || !is_created()){
      OAPR_THROW_LVL(logic_error,
		     "socket is not opened",exception_base::EXC_Crit);
    }
    apr_socket_t *newsock=0;
    basic_io::pool_ptr newpool(new pool_t("socket"));

    /**\bug
     *
     * The following line of code, coupled with pool::operator&,
     * appears to poke a new apr_pool_t* into the pool::m_pool member variable,
     * overwriting the parentless pool created by the constructor. Said pool will
     * not be freed.
     * Correct me if I'm wrong in this analysis, it happens.
     * Giving out non-const references to internal members is bad mmkay :)
     *
     * OAPR_CHECK(apr_pool_create(&(*newpool),m_pool->getp()));
     *
     * -matt
     */

    /** See pool::operator& for why I had to change this line anyway.
      * The bug still exists. */
    OAPR_CHECK(apr_pool_create(newpool->getpp(),m_pool->getp()));
    OAPR_CHECK(apr_socket_accept(&newsock,m_sock,newpool->getp()));
    try{
      basic_socket sock(newsock,newpool,m_socktype);
      return sock;
    }OAPR_CATCH()

  }
  apr_status_t basic_socket::connect(const sockinfo &si)
  {
    if(!is_created()){
      OAPR_THROW_LVL(logic_error,"socket is not created",exception_base::EXC_Error);
    }
    apr_status_t stat;
    stat=apr_socket_connect(m_sock,si.sockaddr());
    if(stat==APR_SUCCESS){
      m_isopen=true;
    }
    return stat;
  }
  apr_status_t basic_socket::option(apr_int32_t opt, apr_int32_t on)
  {
    if(!is_created()){
      OAPR_THROW_LVL(logic_error,
		     "socket is not created",exception_base::EXC_Error);
    }
    return apr_socket_opt_set(m_sock,opt,on);
  }

  apr_status_t basic_socket::close()
  {
    if(is_open()){
      if(apr_socket_shutdown(m_sock,APR_SHUTDOWN_READWRITE)!=APR_SUCCESS){
	m_isopen=false;
	m_created=false;
	apr_socket_close(m_sock);
      }
      m_isopen=false;
      m_created=false;
      return apr_socket_close(m_sock);

    }else{
      return APR_EINVALSOCK;
    }
  }
  bool basic_socket::eof()
  {
    return false;/**\todo is this valid? or should there be a check?*/
  }

  apr_size_t basic_socket::write(const void* buff,
				 apr_size_t toWrite,
				 bool full)
  {
    if(!is_created() || !is_open()){
      OAPR_THROW_LVL(logic_error,
		     "socket is not open",exception_base::EXC_Error);
    }

    apr_size_t sizes(toWrite);
    apr_status_t stat;
    switch(m_socktype){
    case 0:
      OAPR_THROW_LVL(invalid_arg,
		     "invalid socket type",exception_base::EXC_Error);
      break;
    case SOCK_STREAM:
      stat=apr_socket_send(m_sock, static_cast<const char*>(buff), &sizes);
      break;
    case SOCK_DGRAM:
      stat=apr_socket_sendto(m_sock,m_remote,0,static_cast<const char*>(buff),
			     &sizes);
      break;
    default:
       OAPR_THROW_LVL(invalid_arg,
			   "unkown socket type",exception_base::EXC_Error);
       break;
    };
    OAPR_CHECK(stat);
    return sizes;
  }
  apr_size_t basic_socket::read( void* buff,
				 apr_size_t toRead,
				 bool full)const
  {
    if(!is_created() || !is_open()){
      OAPR_THROW_LVL(logic_error,
		     "socket is not open",exception_base::EXC_Error);
    }
    apr_status_t stat;
    apr_size_t sizes(toRead);
    switch(m_socktype){
    case 0:
      OAPR_THROW_LVL(invalid_arg,
		     "invalid socket type",exception_base::EXC_Error);
      break;
    case SOCK_STREAM:
      stat=apr_socket_recv(m_sock, static_cast<char*>(buff),&sizes);
      break;
    case SOCK_DGRAM:
      stat=apr_socket_recvfrom(m_remote,m_sock,0,
			       static_cast<char*>(buff),&sizes);
      break;
    default:
      OAPR_THROW_LVL(invalid_arg,
		     "unknown socket type",exception_base::EXC_Error);
    };

    OAPR_CHECK(stat);

    return sizes;
  }


  apr_status_t basic_socket::flush()
  {
    return 0;
  }

  apr_status_t basic_socket::timeout(apr_interval_time_t timeout)
  {
    if(!is_created()){
      OAPR_THROW_LVL(logic_error,
        "socket not created",exception_base::EXC_Error);
    }
    return apr_socket_timeout_set(m_sock,timeout);
  }
  apr_interval_time_t basic_socket::timeout()const
  {
    if(!is_created()){
      OAPR_THROW_LVL(logic_error,
        "socket not created",exception_base::EXC_Error);
    }
    apr_interval_time_t to;
    OAPR_CHECK(apr_socket_timeout_get(m_sock,&to));
    return to;
  }

};
