/* ====================================================================
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 2003 Matt Carter. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Apache" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 */
#ifndef OAPX_OAPR_SYNCBUFFER
#define OAPX_OAPR_SYNCBUFFER
#include "boost/scoped_array.hpp"
#include "oapr/oapr_mutex.hpp"

namespace oapr
{
  template<typename char_t=char,int sz_t=255>
  class sync_buffer
  {
  public:
    typedef char_t char_type;
  public:
    sync_buffer(const pool::pool_ptr& pl)
      : m_buf(new char_type[sz_t])
	, m_pos(0)
	, m_eptr(0)
	, m_mutex(pl)
	, m_read(pl)
	, m_write(pl)
    {
      m_pos=&m_buf[0];
      m_eptr=&m_buf[sz_t-1];
    }

    int avail_r()const{
      return m_pos-m_buf.get();
    }
    int avail_w()const{
      return m_eptr-m_pos;
    }
    void wait_r()
    {
      oapr::mutex mtx;
      oapr::mlock lk(mtx);
      m_read.wait(mtx);      
    }
    void wait_w()
    {
      oapr::mutex mtx;
      oapr::mlock lk(mtx);
      m_write.wait(mtx);
    }
    int write(const char_t*const buf, int sz)
    {
      oapr::mlock lk(m_mutex);
      /* verify we can put the whole block*/
      int have = m_eptr-m_pos;
      int wr = have>sz ? sz : have;
      std::memcpy(m_pos,buf,wr);
      m_pos+=wr;
      m_write.signal();
      return wr;
    }
    int read(char_t* buf, int sz)const
    {
      oapr::mlock lk(m_mutex);
      int have = m_pos-m_buf.get();
      int rd = have<sz ? have : sz;
      std::memcpy(buf,m_buf.get(),rd);
      std::memmove(m_buf.get(),m_buf.get()+rd,rd);
      m_pos -= rd;
      m_read.signal();
      return rd;
    }
  private:
    sync_buffer(const sync_buffer& sb){}
  private:
    boost::scoped_array<char_type> m_buf;
    mutable char_type *m_pos;
    char_type *m_eptr;
    mutable oapr::mutex m_mutex;
    mutable oapr::condition m_read;
    oapr::condition m_write;
  };
};
#endif
